type Cells = Record<string, {
	value: any,
	addOnChange: (callback: (field: string) => void, field: string) => void,
	removeOnChange: (callback: () => void) => void
}>

type State<T extends Cells> = {
	[K in keyof T]: T[K]["value"]
}

type OnChangeListener<T extends Cells> = (state: State<T>, field: string) => void

export function createStore<T extends Cells>(fields: T) {
	const onChangeListeners: Array<OnChangeListener<T>> = []

	const result = {
		...fields,
		getState() {
			return <State<T>>Object.keys(fields).reduce((state, field) => {
				const cell = fields[field]
				state[field] = cell.value

				return state
			}, {})
		},
		addOnChange(callback) {
			onChangeListeners.push(callback)
		},
		removeOnChange(callback) {
			const index = onChangeListeners.indexOf(callback)

			if (index > -1) {
				onChangeListeners.splice(index, 1)
			}
		}
	} as T & {
		getState: () => State<T>,
		addOnChange: (callback: OnChangeListener<T>) => void,
		removeOnChange: (callback: (state: State<T>) => void) => void
	}

	Object.keys(fields).forEach((field) => {
		fields[field].addOnChange((param) => {
			onChangeListeners.forEach((handler) => handler(result.getState(), param))
		}, field)
	})

	return result
}
